@@ -22,6 +22,7 @@
 ****************************************************************************/
 
 #include "vncview.h"
+#include "krdc_debug.h"
 
 #include <QApplication>
 #include <QImage>
@@ -37,10 +38,10 @@
 #else
     #include "settings.h"
     #include <KActionCollection>
-    #include <KMainWindow>
-    #include <KMessageBox>
-    #include <KPasswordDialog>
-    #include <KXMLGUIClient>
+    #include <KXmlGui/KMainWindow>
+    #include <KWidgetsAddons/KMessageBox>
+    #include <KWidgetsAddons/KPasswordDialog>
+    #include <KXmlGui/KXMLGUIClient>
 #endif
 
 // Definition of key modifier mask constants
@@ -50,7 +51,7 @@
 #define KMOD_Control_L 	0x08
 #define KMOD_Shift_L	0x10
 
-VncView::VncView(QWidget *parent, const KUrl &url, KConfigGroup configGroup)
+VncView::VncView(QWidget *parent, const QUrl &url, KConfigGroup configGroup)
         : RemoteView(parent),
         m_initDone(false),
         m_buttonMask(0),
@@ -61,11 +62,20 @@ VncView::VncView(QWidget *parent, const KUrl &url, KConfigGroup configGroup)
         m_horizontalFactor(1.0),
         m_verticalFactor(1.0),
         m_forceLocalCursor(false)
+#ifdef LIBSSH_FOUND
+        , m_sshTunnelThread(nullptr)
+#endif
 {
     m_url = url;
     m_host = url.host();
     m_port = url.port();
 
+    if (m_port <= 0) // port is invalid or empty...
+        m_port = 5900; // fallback: try an often used VNC port
+
+    if (m_port < 100) // the user most likely used the short form (e.g. :1)
+        m_port += 5900;
+
     // BlockingQueuedConnection can cause deadlocks when exiting, handled in startQuitting()
     connect(&vncThread, SIGNAL(imageUpdated(int,int,int,int)), this, SLOT(updateImage(int,int,int,int)), Qt::BlockingQueuedConnection);
     connect(&vncThread, SIGNAL(gotCut(QString)), this, SLOT(setCut(QString)), Qt::BlockingQueuedConnection);
@@ -74,7 +84,7 @@ VncView::VncView(QWidget *parent, const KUrl &url, KConfigGroup configGroup)
 
     m_clipboard = QApplication::clipboard();
     connect(m_clipboard, SIGNAL(dataChanged()), this, SLOT(clipboardDataChanged()));
-    
+
 #ifndef QTONLY
     m_hostPreferences = new VncHostPreferences(configGroup, this);
 #else
@@ -121,8 +131,8 @@ QSize VncView::minimumSizeHint() const
 void VncView::scaleResize(int w, int h)
 {
     RemoteView::scaleResize(w, h);
-    
-    kDebug(5011) << w << h;
+
+    qCDebug(KRDC) << w << h;
     if (m_scale) {
         m_verticalFactor = (qreal) h / m_frame.height();
         m_horizontalFactor = (qreal) w / m_frame.width();
@@ -139,7 +149,7 @@ void VncView::scaleResize(int w, int h)
         const qreal newH = m_frame.height() * m_verticalFactor;
         setMaximumSize(newW, newH); //This is a hack to force Qt to center the view in the scroll area
         resize(newW, newH);
-    } 
+    }
 }
 
 void VncView::updateConfiguration()
@@ -152,7 +162,7 @@ void VncView::updateConfiguration()
 
 void VncView::startQuitting()
 {
-    kDebug(5011) << "about to quit";
+    qCDebug(KRDC) << "about to quit";
 
     setStatus(Disconnecting);
 
@@ -167,18 +177,25 @@ void VncView::startQuitting()
 
     vncThread.quit();
 
+#ifdef LIBSSH_FOUND
+    if (m_sshTunnelThread) {
+        delete m_sshTunnelThread;
+        m_sshTunnelThread = nullptr;
+    }
+#endif
+
     const bool quitSuccess = vncThread.wait(500);
     if (!quitSuccess) {
         // happens when vncThread wants to call a slot via BlockingQueuedConnection,
         // needs an event loop in this thread so execution continues after 'emit'
         QEventLoop loop;
         if (!loop.processEvents()) {
-            kDebug(5011) << "BUG: deadlocked, but no events to deliver?";
+            qCDebug(KRDC) << "BUG: deadlocked, but no events to deliver?";
         }
         vncThread.wait(500);
     }
 
-    kDebug(5011) << "Quit VNC thread success:" << quitSuccess;
+    qCDebug(KRDC) << "Quit VNC thread success:" << quitSuccess;
 
     setStatus(Disconnected);
 }
@@ -190,8 +207,33 @@ bool VncView::isQuitting()
 
 bool VncView::start()
 {
-    vncThread.setHost(m_host);
-    vncThread.setPort(m_port);
+    QString vncHost = m_host;
+    int vncPort = m_port;
+#ifdef LIBSSH_FOUND
+    if (m_hostPreferences->useSshTunnel()) {
+        Q_ASSERT(!m_sshTunnelThread);
+
+        const int tunnelPort = 58219; // Just a random port
+
+        m_sshTunnelThread = new VncSshTunnelThread(m_host.toUtf8(),
+                                                   m_port,
+                                                   tunnelPort,
+                                                   m_hostPreferences->sshTunnelPort(),
+                                                   m_hostPreferences->sshTunnelUserName().toUtf8(),
+                                                   m_hostPreferences->useSshTunnelLoopback());
+        connect(m_sshTunnelThread, &VncSshTunnelThread::passwordRequest, this, &VncView::sshRequestPassword, Qt::BlockingQueuedConnection);
+        connect(m_sshTunnelThread, &VncSshTunnelThread::errorMessage, this, &VncView::sshErrorMessage);
+        m_sshTunnelThread->start();
+
+        if (m_hostPreferences->useSshTunnelLoopback()) {
+            vncHost = QStringLiteral("127.0.0.1");
+        }
+        vncPort = tunnelPort;
+    }
+#endif
+
+    vncThread.setHost(vncHost);
+    vncThread.setPort(vncPort);
     RemoteView::Quality quality;
 #ifdef QTONLY
     quality = (RemoteView::Quality)((QCoreApplication::arguments().count() > 2) ?
@@ -209,13 +251,21 @@ bool VncView::start()
         // KRDC does always just have one main window, so at(0) is safe
         KXMLGUIClient *mainWindow = dynamic_cast<KXMLGUIClient*>(KMainWindow::memberList().at(0));
         if (mainWindow)
-            mainWindow->actionCollection()->action("show_local_cursor")->setChecked(true);
+            mainWindow->actionCollection()->action(QLatin1String("show_local_cursor"))->setChecked(true);
 #endif
     }
 
     setStatus(Connecting);
 
-    vncThread.start();
+#ifdef LIBSSH_FOUND
+    if (m_hostPreferences->useSshTunnel()) {
+        connect(m_sshTunnelThread, &VncSshTunnelThread::listenReady, this, [this] { vncThread.start(); });
+    }
+    else
+#endif
+    {
+        vncThread.start();
+    }
     return true;
 }
 
@@ -229,9 +279,14 @@ bool VncView::supportsLocalCursor() const
     return true;
 }
 
+bool VncView::supportsViewOnly() const
+{
+    return true;
+}
+
 void VncView::requestPassword(bool includingUsername)
 {
-    kDebug(5011) << "request password";
+    qCDebug(KRDC) << "request password";
 
     setStatus(Authenticating);
 
@@ -240,7 +295,7 @@ void VncView::requestPassword(bool includingUsername)
     }
 
 #ifndef QTONLY
-    // just try to get the passwort from the wallet the first time, otherwise it will loop (see issue #226283)
+    // just try to get the password from the wallet the first time, otherwise it will loop (see issue #226283)
     if (m_firstPasswordTry && m_hostPreferences->walletSupport()) {
         QString walletPassword = readWalletPassword();
 
@@ -290,17 +345,42 @@ void VncView::requestPassword(bool includingUsername)
         vncThread.setPassword(dialog.password());
         if (includingUsername) vncThread.setUsername(dialog.username());
     } else {
-        kDebug(5011) << "password dialog not accepted";
+        qCDebug(KRDC) << "password dialog not accepted";
         startQuitting();
     }
 #endif
 }
 
+#ifdef LIBSSH_FOUND
+void VncView::sshRequestPassword(VncSshTunnelThread::PasswordRequestFlags flags)
+{
+    qCDebug(KRDC) << "request ssh password";
+
+    if (m_hostPreferences->walletSupport() && ((flags & VncSshTunnelThread::IgnoreWallet) != VncSshTunnelThread::IgnoreWallet)) {
+        const QString walletPassword = readWalletSshPassword();
+
+        if (!walletPassword.isNull()) {
+            m_sshTunnelThread->setPassword(walletPassword, VncSshTunnelThread::PasswordFromWallet);
+            return;
+        }
+    }
+
+    KPasswordDialog dialog(this);
+    dialog.setPrompt(i18n("Please enter the SSH password."));
+    if (dialog.exec() == KPasswordDialog::Accepted) {
+        m_sshTunnelThread->setPassword(dialog.password(), VncSshTunnelThread::PasswordFromDialog);
+    } else {
+        qCDebug(KRDC) << "ssh password dialog not accepted";
+        startQuitting();
+    }
+}
+#endif
+
 void VncView::outputErrorMessage(const QString &message)
 {
-    kDebug(5011) << message;
+    qCritical(KRDC) << message;
 
-    if (message == "INTERNAL:APPLE_VNC_COMPATIBILTY") {
+    if (message == QLatin1String("INTERNAL:APPLE_VNC_COMPATIBILTY")) {
         setCursor(localDotCursor());
         m_forceLocalCursor = true;
         return;
@@ -313,6 +393,17 @@ void VncView::outputErrorMessage(const QString &message)
     emit errorMessage(i18n("VNC failure"), message);
 }
 
+void VncView::sshErrorMessage(const QString &message)
+{
+    qCritical(KRDC) << message;
+
+    startQuitting();
+
+    KMessageBox::error(this, message, i18n("SSH Tunnel failure"));
+
+    emit errorMessage(i18n("SSH Tunnel failure"), message);
+}
+
 #ifndef QTONLY
 HostPreferences* VncView::hostPreferences()
 {
@@ -322,7 +413,7 @@ HostPreferences* VncView::hostPreferences()
 
 void VncView::updateImage(int x, int y, int w, int h)
 {
-//     kDebug(5011) << "got update" << width() << height();
+//     qCDebug(KRDC) << "got update" << width() << height();
 
     m_x = x;
     m_y = y;
@@ -353,13 +444,13 @@ void VncView::updateImage(int x, int y, int w, int h)
         setFocusPolicy(Qt::WheelFocus);
         setStatus(Connected);
         emit connected();
-        
+
         if (m_scale) {
 #ifndef QTONLY
-            kDebug(5011) << "Setting initial size w:" <<m_hostPreferences->width() << " h:" << m_hostPreferences->height();
+            qCDebug(KRDC) << "Setting initial size w:" <<m_hostPreferences->width() << " h:" << m_hostPreferences->height();
             emit framebufferSizeChanged(m_hostPreferences->width(), m_hostPreferences->height());
             scaleResize(m_hostPreferences->width(), m_hostPreferences->height());
-            kDebug() << "m_frame.size():" << m_frame.size() << "size()" << size();
+            qCDebug(KRDC) << "m_frame.size():" << m_frame.size() << "size()" << size();
 #else
 //TODO: qtonly alternative
 #endif
@@ -370,18 +461,23 @@ void VncView::updateImage(int x, int y, int w, int h)
 #ifndef QTONLY
         if (m_hostPreferences->walletSupport()) {
             saveWalletPassword(vncThread.password());
+#ifdef LIBSSH_FOUND
+            if (m_hostPreferences->useSshTunnel()) {
+                saveWalletSshPassword();
+            }
+#endif
         }
 #endif
     }
 
     if ((y == 0 && x == 0) && (m_frame.size() != size())) {
-        kDebug(5011) << "Updating framebuffer size";
+        qCDebug(KRDC) << "Updating framebuffer size";
         if (m_scale) {
             setMaximumSize(QSize(QWIDGETSIZE_MAX, QWIDGETSIZE_MAX));
             if (parentWidget())
                 scaleResize(parentWidget()->width(), parentWidget()->height());
         } else {
-            kDebug(5011) << "Resizing: " << m_frame.width() << m_frame.height();
+            qCDebug(KRDC) << "Resizing: " << m_frame.width() << m_frame.height();
             resize(m_frame.width(), m_frame.height());
             setMaximumSize(m_frame.width(), m_frame.height()); //This is a hack to force Qt to center the view in the scroll area
             setMinimumSize(m_frame.width(), m_frame.height());
@@ -441,9 +537,9 @@ void VncView::setCut(const QString &text)
 
 void VncView::paintEvent(QPaintEvent *event)
 {
-//     kDebug(5011) << "paint event: x: " << m_x << ", y: " << m_y << ", w: " << m_w << ", h: " << m_h;
+//     qCDebug(KRDC) << "paint event: x: " << m_x << ", y: " << m_y << ", w: " << m_w << ", h: " << m_h;
     if (m_frame.isNull() || m_frame.format() == QImage::Format_Invalid) {
-        kDebug(5011) << "no valid image to paint";
+        qCDebug(KRDC) << "no valid image to paint";
         RemoteView::paintEvent(event);
         return;
     }
@@ -453,27 +549,27 @@ void VncView::paintEvent(QPaintEvent *event)
     QPainter painter(this);
 
     if (m_repaint) {
-//         kDebug(5011) << "normal repaint";
+//         qCDebug(KRDC) << "normal repaint";
         painter.drawImage(QRect(qRound(m_x*m_horizontalFactor), qRound(m_y*m_verticalFactor),
-                                qRound(m_w*m_horizontalFactor), qRound(m_h*m_verticalFactor)), 
-                          m_frame.copy(m_x, m_y, m_w, m_h).scaled(qRound(m_w*m_horizontalFactor), 
+                                qRound(m_w*m_horizontalFactor), qRound(m_h*m_verticalFactor)),
+                          m_frame.copy(m_x, m_y, m_w, m_h).scaled(qRound(m_w*m_horizontalFactor),
                                                                   qRound(m_h*m_verticalFactor),
                                                                   Qt::IgnoreAspectRatio, Qt::SmoothTransformation));
     } else {
-//         kDebug(5011) << "resize repaint";
+//         qCDebug(KRDC) << "resize repaint";
         QRect rect = event->rect();
         if (rect.width() != width() || rect.height() != height()) {
-//             kDebug(5011) << "Partial repaint";
+//             qCDebug(KRDC) << "Partial repaint";
             const int sx = rect.x()/m_horizontalFactor;
             const int sy = rect.y()/m_verticalFactor;
             const int sw = rect.width()/m_horizontalFactor;
             const int sh = rect.height()/m_verticalFactor;
-            painter.drawImage(rect, 
+            painter.drawImage(rect,
                               m_frame.copy(sx, sy, sw, sh).scaled(rect.width(), rect.height(),
                                                                   Qt::IgnoreAspectRatio, Qt::SmoothTransformation));
         } else {
-//             kDebug(5011) << "Full repaint" << width() << height() << m_frame.width() << m_frame.height();
-            painter.drawImage(QRect(0, 0, width(), height()), 
+//             qCDebug(KRDC) << "Full repaint" << width() << height() << m_frame.width() << m_frame.height();
+            painter.drawImage(QRect(0, 0, width(), height()),
                               m_frame.scaled(m_frame.width() * m_horizontalFactor, m_frame.height() * m_verticalFactor,
                                              Qt::IgnoreAspectRatio, Qt::SmoothTransformation));
         }
@@ -493,7 +589,7 @@ bool VncView::event(QEvent *event)
     switch (event->type()) {
     case QEvent::KeyPress:
     case QEvent::KeyRelease:
-//         kDebug(5011) << "keyEvent";
+//         qCDebug(KRDC) << "keyEvent";
         keyEventHandler(static_cast<QKeyEvent*>(event));
         return true;
         break;
@@ -501,12 +597,12 @@ bool VncView::event(QEvent *event)
     case QEvent::MouseButtonPress:
     case QEvent::MouseButtonRelease:
     case QEvent::MouseMove:
-//         kDebug(5011) << "mouseEvent";
+//         qCDebug(KRDC) << "mouseEvent";
         mouseEventHandler(static_cast<QMouseEvent*>(event));
         return true;
         break;
     case QEvent::Wheel:
-//         kDebug(5011) << "wheelEvent";
+//         qCDebug(KRDC) << "wheelEvent";
         wheelEventHandler(static_cast<QWheelEvent*>(event));
         return true;
         break;
@@ -554,6 +650,18 @@ void VncView::wheelEventHandler(QWheelEvent *event)
     vncThread.mouseEvent(x, y, m_buttonMask);
 }
 
+#ifdef LIBSSH_FOUND
+QString VncView::readWalletSshPassword()
+{
+    return readWalletPasswordForKey(QStringLiteral("SSHTUNNEL") + m_url.toDisplayString(QUrl::StripTrailingSlash));
+}
+
+void VncView::saveWalletSshPassword()
+{
+    saveWalletPasswordForKey(QStringLiteral("SSHTUNNEL") + m_url.toDisplayString(QUrl::StripTrailingSlash), m_sshTunnelThread->password());
+}
+#endif
+
 void VncView::keyEventHandler(QKeyEvent *e)
 {
     // strip away autorepeating KeyRelease; see bug #206598
@@ -600,8 +708,6 @@ void VncView::unpressModifiers()
 
 void VncView::clipboardDataChanged()
 {
-    kDebug(5011);
-
     if (m_status != Connected)
         return;
 
